﻿using System;
using System.Net;
using System.IO;
using System.Text;
using System.Threading;
using System.Drawing;
using System.Windows.Forms;
using WatcherCommons.Interfaces;
using NetWaveProtocol.Properties;

namespace NetWaveProtocol
{
    public class NetWaveCamera : IWatcherCamera
    {
        private CameraConfiguration _dlgCfg = new CameraConfiguration();
        private Thread _tCapture = null;
        // La resolución es 8 para 320x200 y 32 para 640x480
        // The resolution is 8 for 320x200 and 32 for 640x480
        private int _nResolution = 32;
        private bool _bCapturing = false;
        private bool _nChangeResolution = true;

        public NetWaveCamera()
        {
            _dlgCfg.Camera = this;
        }
        #region IWatcherCamera
        public event NewFrameEventHandler OnNewFrame = null;
        public Size FrameSize
        {
            get
            {
                return _nResolution == 8 ? new Size(320, 200) : new Size(640, 480);
            }
        }
        public string ConnectionString
        {
            get
            {
                return string.Join(";", new string[] { "url=" + Uri, "userName=" + UserName, "password=" + Password });
            }
            set
            {
                foreach (string p in value.Split(';'))
                {
                    string[] part = p.Split('=');
                    switch (part[0].ToLower())
                    {
                        case "url":
                            Uri = part[1];
                            break;
                        case "username":
                            UserName = part[1];
                            break;
                        case "password":
                            Password = part[1];
                            break;
                    }
                }
            }
        }
        public string UserName { get; set; }
        public string Password { get; set; }
        public string Uri { get; set; }
        public int MaxFPS { get; set; }
        public bool Status
        {
            get
            {
                return _bCapturing;
            }
        }
        public ICameraSetupManager SetupManager
        {
            get
            {
                return _dlgCfg;
            }
        }
        public void Initialize()
        {
            Close();
            OnNewFrame = null;
        }
        public void ShowCameraConfiguration(Form parent)
        {
            _dlgCfg.URL = Uri;
            _dlgCfg.UserName = UserName;
            _dlgCfg.Password = Password;
            // Este cuadro de diálogo es no modal, ya que el usuario debe poder ver el resultado de los cambios
            // This dialog box is not modal, since the user should be able to see the result of changes
            _dlgCfg.Show();
            if (_nResolution != _dlgCfg.Resolution)
            {
                _nResolution = _dlgCfg.Resolution;
                _nChangeResolution = true;
            }
        }
        public void Start()
        {
            if (_nChangeResolution)
            {
                string userpwd = UserName + ":" + Password;
                UTF8Encoding enc = new UTF8Encoding();
                // Con este comando fijamos la resolución de la cámara
                // With this command we set the camera resolution
                string url = Uri + (Uri.EndsWith("/") ? "" : "/") + "camera_control.cgi?param=0&value=" + _nResolution.ToString();
                // Utilizamos un HttpWebRequest para enviar y recibir mensajes http a la cámara por
                // We use an HttpWebRequest to send and receive http messages to the camera
                HttpWebRequest wrq = (HttpWebRequest)HttpWebRequest.Create(url);
                wrq.Method = "GET";
                // Cabecera http para autorizar la petición
                // Http header to authorize the request
                wrq.Headers.Add("Authorization", "Basic " + Convert.ToBase64String(enc.GetBytes(userpwd)));
                // Enviamos la petición
                // Sending the request
                WebResponse wr = wrq.GetResponse();
                wr.Close();
                _nChangeResolution = false;
            }
            // Ponemos en marcha el timer para capturar las imágenes
            // We start the timer to capture the images
            _bCapturing = true;
            _tCapture = new Thread(new ParameterizedThreadStart(CaptureThread));
            _tCapture.Start(MaxFPS);
        }
        public void Close()
        {
            try
            {
                if (_tCapture != null)
                {
                    _bCapturing = false;
                    _tCapture.Abort();
                    _tCapture = null;
                }
            }
            catch
            {
            }
        }
        #endregion
        private void CaptureThread(object interval)
        {
            while (_bCapturing)
            {
                try
                {
                    DateTime tm = DateTime.Now;
                    string userpwd = UserName + ":" + Password;
                    UTF8Encoding enc = new UTF8Encoding();
                    // Comando para capturar una imágen
                    // Command to capture an image
                    string url = Uri + (Uri.EndsWith("/") ? "" : "/") + "snapshot.cgi?user=" + UserName + "&pwd=" + Password;
                    HttpWebRequest wrq = (HttpWebRequest)HttpWebRequest.Create(url);
                    wrq.Method = "GET";
                    wrq.Headers.Add("Authorization", "Basic " + Convert.ToBase64String(enc.GetBytes(userpwd)));
                    WebResponse wr = wrq.GetResponse();
                    // Recibimos un stream con una imagen en formato JPG
                    // We receive a stream with an image in JPG format
                    Stream receiveStream = wr.GetResponseStream();
                    // Y la convertimos en un bitmap
                    // Por último, la pasamos al programa para que la procese
                    // And we turn it into a bitmap
                    // Finally, we provide the image to the program to be processed
                    OnNewFrame?.Invoke(this, new FrameEventArgs(new Bitmap(receiveStream)));
                    wr.Close();
                    TimeSpan ts;
                    do
                    {
                        ts = DateTime.Now.Subtract(tm);
                    } while (((ts.Seconds * 1000) + ts.Milliseconds) <= (int)interval);
                }
                catch
                {
                }
            }
        }
        public void SetResolution(int res)
        {
            if ((res == 8) || (res == 32))
            {
                _nResolution = res;
                _nChangeResolution = true;
            }
        }
        public override string ToString()
        {
            // Texto que se mostrará en la lista de cámaras
            // Text to be displayed in the camera list
            return Resources.ID_PROTOCOL;
        }
    }
}
