﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Configuration;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;
using WatcherCommons.Interfaces;
using WatcherCommons.Data;
using ThiefWatcher.Properties;
using ThiefWatcher.Configuration;

namespace ThiefWatcher
{
    public enum CameraMode
    {
        Configuration, Surveillance, Simulacrum, Active
    }
    public partial class frmCamera : Form
    {
        private IWatcherCamera _camera = null;
        private CameraMode _mode = CameraMode.Configuration;
        private string _url = "";
        private bool _takePhoto = false;
        private DateTime _nextPhoto;
        private int _photoCnt = 0;
        private List<string> _clients = new List<string>();
        private Semaphore _sClients = new Semaphore(1, 1);
        public frmCamera()
        {
            InitializeComponent();
            Translate();
        }
        public frmCamera(CameraData camera)
        {
            InitializeComponent();
            Translate();
            ProtocolsSection protocols = (ProtocolsSection)ConfigurationManager.GetSection(Resources.STR_PROTOCOLSECTION);
            _camera = (IWatcherCamera)Type.GetType(protocols.Protocols.GetByName(camera.Protocol).ProtocolType).GetConstructor(Type.EmptyTypes).Invoke(null);
            _camera.ConnectionString = ConfigurationManager.ConnectionStrings[camera.ConnectionString].ConnectionString;
            _camera.OnNewFrame += new NewFrameEventHandler(OnNewImage);
            _camera.SetupManager.OnFrameSizeChanged += new EventHandler(OnChangeImageSize);            
            _url = _camera.Uri;
            txtID.Text = camera.ID;
            ComposeTitle();
            bSettings.Enabled = true;
            bStart.Enabled = true;
        }
        public IWatcherCamera Camera
        {
            get
            {
                return _camera;
            }
            set
            {
                if (_camera != null)
                {
                    _camera.Close();
                    pbImage.Image = null;
                }
                _camera = value;
                bIdentify_Click(null, EventArgs.Empty);
                if (_camera != null)
                {
                    _camera.OnNewFrame += new NewFrameEventHandler(OnNewImage);
                    _camera.SetupManager.OnFrameSizeChanged += new EventHandler(OnChangeImageSize);
                    _camera.ShowCameraConfiguration(this);
                }
            }
        }
        public IStorageManager StorageManager { get; set; }
        public string ID
        {
            get
            {
                return txtID.Text;
            }
            set
            {
                txtID.Text = value;
                ComposeTitle();
            }
        }
        public bool Active
        {
            get
            {
                return _camera.Status;
            }
        }
        public int PhotoCount { get; set; }
        public TimeSpan PhotoTS { get; set; }
        public CameraMode Mode
        {
            get
            {
                return _mode;
            }
            set
            {
                _mode = value;
                ComposeTitle();
                switch (_mode)
                {
                    case CameraMode.Simulacrum:
                        bStart_Click(null, EventArgs.Empty);
                        break;
                    case CameraMode.Surveillance:
                        _takePhoto = true;
                        _photoCnt = 0;
                        _nextPhoto = DateTime.Now.Add(PhotoTS);
                        Start();
                        break;
                    case CameraMode.Active:
                        Start();
                        break;
                    case CameraMode.Configuration:
                        _photoCnt = 0;
                        break;
                }
            }
        }
        public void ClearClients()
        {
            try
            {
                _sClients.WaitOne();
                _clients.Clear();
            }
            finally
            {
                _sClients.Release();
            }
        }
        public void AddClient(string id)
        {
            try
            {
                _sClients.WaitOne();
                if (!_clients.Contains(id))
                {
                    _clients.Add(id);
                }
            }
            finally
            {
                _sClients.Release();
            }
        }
        public void RemoveClient(string id)
        {
            try
            {
                _sClients.WaitOne();
                if (_clients.Contains(id))
                {
                    _clients.Remove(id);
                }
            }
            finally
            {
                _sClients.Release();
            }
        }
        public CameraInfo CameraRequest(CameraInfo req, CameraMode actualmode)
        {
            if (req.ID == ID)
            {
                if (req.Photo)
                {
                    _takePhoto = true;
                }
                else
                {
                    if (req.Active)
                    {
                        if (!Active)
                        {
                            Mode = actualmode;
                        }
                    }
                    else
                    {
                        Stop();
                    }
                }
                req.Active = Active;
                req.Width = Camera.FrameSize.Width;
                req.Height = Camera.FrameSize.Height;
                return req;
            }
            return null;
        }
        public void Stop()
        {
            if (_camera != null)
            {
                try
                {
                    _camera.Close();
                    bStart.Enabled = true;
                    bStop.Enabled = false;
                }
                catch
                {
                    bStart.Enabled = false;
                    bStop.Enabled = true;
                    throw;
                }
            }
        }
        public void Start()
        {
            if (_camera != null)
            {
                try
                {
                    _camera.Start();
                    bStart.Enabled = false;
                    bStop.Enabled = true;
                }
                catch
                {
                    bStart.Enabled = true;
                    bStop.Enabled = false;
                }
            }
        }
        private void Translate()
        {
            bIdentify.ToolTipText = UIResources.TTP_IDENTIFY;
            bSettings.ToolTipText = UIResources.TTP_SETTINGS;
            bStart.ToolTipText = UIResources.TTP_STARTCAMERA;
            bStop.ToolTipText = UIResources.TTP_STOPCAMERA;
            bSave.ToolTipText = UIResources.TTP_SAVECFG;
            bDelete.ToolTipText = UIResources.TTP_REMOVECFG;
        }
        private void ComposeTitle()
        {
            string md = UIResources.CM_CONFIG;
            switch (_mode)
            {
                case CameraMode.Active:
                    md = UIResources.CM_ACTIVE;
                    break;
                case CameraMode.Surveillance:
                    md = UIResources.CM_SURV;
                    break;
                case CameraMode.Simulacrum:
                    md = UIResources.CM_SIMULACRUM;
                    break;
            }
            Text = string.Join(" - ", new string[] { ID, _url, md });
        }
        private void SaveFrame(string filename, Bitmap bmp)
        {
            if (StorageManager != null)
            {
                try
                {
                    _sClients.WaitOne();
                    foreach (string id in _clients)
                    {
                        if (!StorageManager.ExistsFile(string.Format(filename, id)))
                        {
                            using (MemoryStream stream = new MemoryStream())
                            {
                                bmp.Save(stream, ImageFormat.Jpeg);
                                stream.Position = 0;
                                StorageManager.UploadFile(string.Format(filename, id), stream);
                            }
                        }
                    }
                }
                finally
                {
                    _sClients.Release();
                }
            }
        }
        private void SavePhoto(string filename, Bitmap bmp)
        {
            if (StorageManager != null)
            {
                if (!StorageManager.ExistsFile(filename))
                {
                    using (MemoryStream stream = new MemoryStream())
                    {
                        bmp.Save(stream, ImageFormat.Jpeg);
                        stream.Position = 0;
                        StorageManager.UploadFile(filename, stream);
                    }
                }
            }
        }
        private void OnNewImage(object sender, FrameEventArgs e)
        {
            try
            {
                if ((_mode == CameraMode.Surveillance) && (_photoCnt < PhotoCount) && (DateTime.Now.CompareTo(_nextPhoto) >= 0))
                {
                    SavePhoto(ID + "_PHOTO_" + DateTime.Now.ToString("yyyyMMddHHmmss") + ".jpg", e.Image);
                    _nextPhoto = _nextPhoto.Add(PhotoTS);
                    _photoCnt++;
                    _takePhoto = false;
                }
                switch (_mode)
                {
                    case CameraMode.Configuration:
                        BeginInvoke((Action)delegate () { pbImage.Image = e.Image; });
                        break;
                    case CameraMode.Surveillance:
                        SaveFrame(ID + "_FRAME_{0}.jpg", e.Image);
                        break;
                    case CameraMode.Simulacrum:
                        SaveFrame(ID + "_FRAME_{0}.jpg", e.Image);
                        IAsyncResult ir = BeginInvoke((Action)delegate () { pbImage.Image = e.Image; });
                        while (!ir.IsCompleted) ;
                        break;
                    case CameraMode.Active:
                        SaveFrame(ID + "_FRAME_{0}.jpg", e.Image);
                        break;
                }
                if (_takePhoto)
                {
                    SavePhoto(ID + "_PHOTO_" + DateTime.Now.ToString("yyyyMMddHHmmss") + ".jpg", e.Image);
                    _takePhoto = false;
                }
            }
            catch (Exception ex)
            {
                switch (_mode)
                {
                    case CameraMode.Configuration:
                        MessageBox.Show(ex.Message);
                        break;
                }
            }       
        }
        private void OnChangeImageSize(object sender, EventArgs e)
        {
            try
            {
                SetClientSizeCore(_camera.FrameSize.Width, _camera.FrameSize.Height);
            }
            catch { }
        }

        private void bIdentify_Click(object sender, EventArgs e)
        {
            if (_camera != null)
            {
                try
                {
                    AuthenticationParameters dlg = new AuthenticationParameters();
                    dlg.UserName = _camera.UserName;
                    dlg.Password = _camera.Password;
                    dlg.URL = _camera.Uri;
                    if (dlg.ShowDialog(this) == DialogResult.OK)
                    {
                        _camera.UserName = dlg.UserName;
                        _camera.Password = dlg.Password;
                        _camera.Uri = dlg.URL;
                        _url = dlg.URL;
                        ComposeTitle();
                        bSettings.Enabled = true;
                        bStart.Enabled = true;
                        if (sender != null)
                        {
                            _camera.SetupManager.ReloadSettings();
                        }
                        return;
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
            bSettings.Enabled = false;
            bStart.Enabled = false;
            bStop.Enabled = false;
        }

        private void bSettings_Click(object sender, EventArgs e)
        {
            if (_camera != null)
            {
                try
                {
                    _camera.ShowCameraConfiguration(this);

                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void bStart_Click(object sender, EventArgs e)
        {
            if (_camera != null)
            {
                try
                {
                    SetClientSizeCore(_camera.FrameSize.Width, _camera.FrameSize.Height);
                    _camera.Start();
                    bStart.Enabled = false;
                    bStop.Enabled = true;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                    bStart.Enabled = true;
                    bStop.Enabled = false;
                }
            }
        }

        private void bStop_Click(object sender, EventArgs e)
        {
            if (_camera != null)
            {
                try
                {
                    _camera.Close();
                    bStart.Enabled = true;
                    bStop.Enabled = false;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                    bStart.Enabled = false;
                    bStop.Enabled = true;
                }
            }
        }
        private void frmCamera_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (_camera != null)
            {
                try
                {
                    _camera.Close();
                }
                catch (Exception ex)
                {
                    e.Cancel = true;
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void txtID_TextChanged(object sender, EventArgs e)
        {
            ComposeTitle();
        }

        private void bSave_Click(object sender, EventArgs e)
        {
            try
            {
                if (string.IsNullOrEmpty(txtID.Text))
                {
                    throw new Exception(UIResources.ERR_NOID);
                }
                ConfigurationHelper.SaveConnectionString(txtID.Text, _camera.ConnectionString, true);
                string protocolname = ConfigurationHelper.GetProtocol(_camera.GetType());
                System.Configuration.Configuration con = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
                CamerasSection cameras = (CamerasSection)con.GetSection(Resources.STR_CAMSECTION);
                bool saved = false;
                for (int ix = 0; ix < cameras.Cameras.Count; ix++)
                {
                    if (string.Compare(cameras.Cameras[ix].ID, txtID.Text, true) == 0)
                    {
                        cameras.Cameras[ix].Protocol = protocolname;
                        cameras.Cameras[ix].ConnectionString = txtID.Text;
                        saved = true;             
                    }
                }
                if (!saved)
                {
                    CameraData camcfg = new CameraData();
                    camcfg.ID = txtID.Text;
                    camcfg.Protocol = protocolname;
                    camcfg.ConnectionString = txtID.Text;
                    cameras.Cameras.Add(camcfg);
                }
                con.Save(ConfigurationSaveMode.Modified);
                ConfigurationManager.RefreshSection(Resources.STR_CAMSECTION);
                MessageBox.Show(UIResources.MSG_CFGSAVED);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void bDelete_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show(UIResources.MSG_DELCFG, UIResources.CAP_WARNING, MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.Yes)
            {
                try
                {
                    System.Configuration.Configuration con = ConfigurationManager.OpenExeConfiguration(ConfigurationUserLevel.None);
                    CamerasSection cameras = (CamerasSection)con.GetSection(Resources.STR_CAMSECTION);
                    for (int ix = 0; ix < cameras.Cameras.Count; ix++)
                    {
                        if (string.Compare(cameras.Cameras[ix].ID, txtID.Text, true) == 0)
                        {
                            cameras.Cameras.RemoveAt(ix);
                            break;
                        }
                    }
                    con.Save(ConfigurationSaveMode.Modified);
                    ConfigurationManager.RefreshSection(Resources.STR_CAMSECTION);
                    MessageBox.Show(UIResources.MSG_CFGSAVED);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
             }
        }
    }
}
