﻿using System;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.IO;
using System.Xml;
using VAPIXProtocol.Properties;
using VAPIXProtocol.Commands;
using WatcherCommons.Interfaces;

namespace VAPIXProtocol
{
    // Proporciona soporte para configurar los parámetros de la cámara, como brillo, contraste y movimiento
    // Provides support to configure camera settings, such as brightness, contrast and movement
    public partial class CameraConfiguration : Form, ICameraSetupManager
    {
        public CameraConfiguration()
        {
            InitializeComponent();
            Text = UIResources.TTL_CAMERACONTROL;
            tpParams.Text = UIResources.LAB_PARAMS;
        }
        public event EventHandler OnFrameSizeChanged = null;
        public void ReloadSettings()
        {
            try
            {
                Cursor = Cursors.WaitCursor;
                GetCurrentSettings();
                GetPTZCommands();
                Cursor = Cursors.Default;
            }
            catch (Exception ex)
            {
                Cursor = Cursors.Default;
                MessageBox.Show(ex.Message);
            }
        }
        public string URL { get; set; }
        public string UserName { get; set; }
        public string Password { get; set; }
        public string CamWidth { get; set; }
        public string CamHeight { get; set; }
        public VAPIXCamera Camera { get; set; }
        
        // Soporte para comados cgi que devuelven texto como respuesta
        // Support for cgi commands that return text as response
        private string RequestCommand(string url, int timeout = 2000)
        {
            HttpWebRequest wrq = (HttpWebRequest)HttpWebRequest.Create(url);
            wrq.Method = "GET";
            // Proprciona un timeot corto a las operaciones de control para dar prioridad a la captura de imágenes
            // Provide a short timeot to the control operations to priorize the image capture
            wrq.Timeout = timeout;
            // Autenticación básica por encabezados http
            // Http headers basic Authentication
            wrq.Credentials = new NetworkCredential(UserName, Password);
            string response;
            WebResponse wr = wrq.GetResponse();
            Stream receiveStream = wr.GetResponseStream();
            StreamReader reader = new StreamReader(receiveStream, Encoding.UTF8);
            response = reader.ReadToEnd();
            wr.Close();
            return response;
        }

        private void AddParameter(string name, string nicename, string path, TreeNodeCollection nodes, XmlElement ge)
        {
            if (!nodes.ContainsKey(name))
            {
                TreeNode node = nodes.Add(name, string.IsNullOrEmpty(nicename) ? name : nicename);
                if (string.Compare(ge.Name, "group", true) == 0)
                {
                    foreach (XmlElement gr in ge.ChildNodes)
                    {
                        AddParameter(gr.GetAttribute("name"), gr.GetAttribute("niceName"), path + "." + gr.GetAttribute("name"), node.Nodes, gr);
                    }
                }
                else if (string.Compare(ge.Name, "parameter", true) == 0)
                {
                    ParamDefinition pd = new ParamDefinition(ge) { Node = node, FullPath = path };
                    if (!pd.ReadOnly)
                    {
                        pd.OnSave += new SaveParameter(SaveParameterValue);
                    }
                    if (pd.FullPath.EndsWith("Image.I0.Appearance.Resolution"))
                    {
                        string[] reso = pd.Value.Split('x');
                        CamWidth = reso[0];
                        CamHeight = reso[1];
                        OnFrameSizeChanged?.Invoke(Camera, EventArgs.Empty);
                    }
                    node.Tag = pd;
                    node.Text = node.Text + " (" + pd.Value + ")";
                }
            }
        }
        // Obtiene los valores actuales de brillo, contraste y posición de pantalla
        // Gets the current settings of brightness, contrast and screen position
        private void GetCurrentSettings()
        {
            try
            {
                tvParameters.Nodes.Clear();             
                // Otra forma de autenticar la petición es mediante parámetros
                // Another way to authenticate the request is by means of parameters
                string url = URL + (URL.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=listdefinitions&listformat=xmlschema";
                string response = RequestCommand(url);
                XmlDocument doc = new XmlDocument();
                doc.LoadXml(response);
                XmlElement root = doc.DocumentElement;
                foreach(XmlElement ep in root.ChildNodes)
                {
                    if (string.Compare(ep.Name, "group", true) == 0)
                    {
                        AddParameter(ep.GetAttribute("name"), ep.GetAttribute("niceName"), ep.GetAttribute("name"), tvParameters.Nodes, ep);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void AddCommand(string cmd)
        {
            CmdBase cb = CmdBase.CreateCommand(cmd, tvParameters);
            if (cb != null)
            {
                cb.OnExecute += new ExecuteCommand(RunCommand);
                if (!tvPTZ.Nodes.ContainsKey(cb.Command))
                {
                    TreeNode node = tvPTZ.Nodes.Add(cb.Command, cb.Command);
                    node.Tag = cb;
                }
            }
        }
        private void GetPTZCommands()
        {
            try
            {
                string url = URL + (URL.EndsWith("/") ? "" : "/") + "axis-cgi/com/ptz.cgi?info=1&camera=1";
                string response = RequestCommand(url);
                StringReader srd = new StringReader(response);
                if (response.Contains("Error"))
                {
                    throw new Exception(response);
                }
                tvPTZ.Nodes.Clear();
                srd.ReadLine();
                srd.ReadLine();
                srd.ReadLine();
                srd.ReadLine();
                string cmd = srd.ReadLine();
                while (!string.IsNullOrEmpty(cmd))
                {
                    if (!cmd.StartsWith(" "))
                    {
                        AddCommand(cmd);
                    }
                    cmd = srd.ReadLine();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void SaveParameterValue(ParamDefinition pd)
        {
            try
            {
                string url = URL + (URL.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=update&" + pd.FullPath + "=" + pd.Value;
                string response = RequestCommand(url);
                if (response.Contains("Error"))
                {
                    throw new Exception(response);
                }
                pd.Node.Text = pd.NiceName + " (" + pd.Value + ")";
                if (pd.FullPath.EndsWith("Image.I0.Appearance.Resolution"))
                {
                    string[] reso = pd.Value.Split('x');
                    CamWidth = reso[0];
                    CamHeight = reso[1];
                    OnFrameSizeChanged?.Invoke(Camera, EventArgs.Empty);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
                string url = URL + (URL.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=list&group=" + pd.FullPath;
                string[] response = RequestCommand(url).Split('=');
                pd.Value = response[1];
            }
        }

        private void RunCommand(CmdBase cb)
        {
            try
            {
                string url = URL + (URL.EndsWith("/") ? "" : "/") + cb.GetCommandUri(pValues) + "&camera=1";
                string response = RequestCommand(url, 4000);
                if (response.Contains("Error"))
                {
                    throw new Exception(response);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void CameraConfiguration_Load(object sender, EventArgs e)
        {
            try
            {
                Cursor = Cursors.WaitCursor;
                GetCurrentSettings();
                Cursor = Cursors.Default;
            }
            catch (Exception ex)
            {
                Cursor = Cursors.Default;
                MessageBox.Show(ex.Message);
            }
        }

        private void CameraConfiguration_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = true;
            Hide();
        }

        private void tvParameters_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node != null)
            {
                pValues.Controls.Clear();
                ParamDefinition pd = e.Node.Tag as ParamDefinition;
                if (pd != null)
                {
                    pd.CreateEditor(pValues);
                }
            }
        }

        private void tcConfig_SelectedIndexChanged(object sender, EventArgs e)
        {
            pValues.Controls.Clear();
            if (tcConfig.SelectedTab.Name == "tpPTZ")
            {
                GetPTZCommands();
            }
        }

        private void tvPTZ_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node != null)
            {
                pValues.Controls.Clear();
                CmdBase cb = e.Node.Tag as CmdBase;
                if (cb != null)
                {
                    cb.CreateUI(pValues);
                }
            }
        }
    }
}
