﻿using System;
using System.Net;
using System.IO;
using System.Text;
using System.Threading;
using System.Drawing;
using System.Windows.Forms;
using WatcherCommons.Interfaces;
using VAPIXProtocol.Properties;

namespace VAPIXProtocol
{
    public class VAPIXCamera : IWatcherCamera
    {
        private CameraConfiguration _dlgCfg = new CameraConfiguration();
        private Thread _tCapture = null;
        private int _nTimerPeriod = 100;
        private bool _nChangeResolution = true;
        private string _width = "640";
        private string _height = "480";

        private bool _bCapturing = false;

        public VAPIXCamera()
        {
            _dlgCfg.Camera = this;            
        }

        #region IWatcherCamera
        public event NewFrameEventHandler OnNewFrame = null;
        public Size FrameSize
        {
            get
            {
                return new Size(int.Parse(_dlgCfg.CamWidth), int.Parse(_dlgCfg.CamHeight));
            }
        }
        public string ConnectionString
        {
            get
            {
                return string.Join(";", new string[] { "url=" + Uri, "userName=" + UserName, "password=" + Password });
            }
            set
            {
                foreach (string p in value.Split(';'))
                {
                    string[] part = p.Split('=');
                    switch (part[0].ToLower())
                    {
                        case "url":
                            Uri = part[1];
                            _dlgCfg.URL = Uri;
                            break;
                        case "username":
                            UserName = part[1];
                            _dlgCfg.UserName = UserName;
                            break;
                        case "password":
                            Password = part[1];
                            _dlgCfg.Password = Password;
                            break;
                    }
                }
                _dlgCfg.ReloadSettings();
            }
        }
        public string UserName { get; set; }
        public string Password { get; set; }
        public string Uri { get; set; }
        public int MaxFPS { get; set; }
        public bool Status
        {
            get
            {
                return _bCapturing;
            }
        }
        public ICameraSetupManager SetupManager
        {
            get
            {
                return _dlgCfg;
            }
        }
        public void Initialize()
        {
            Close();
            OnNewFrame = null;
        }
        public void ShowCameraConfiguration(Form parent)
        {
            _dlgCfg.URL = Uri;
            _dlgCfg.UserName = UserName;
            _dlgCfg.Password = Password;            
            // Este cuadro de diálogo es no modal, ya que el usuario debe poder ver el resultado de los cambios
            // This dialog box is not modal, since the user should be able to see the result of changes
            _dlgCfg.Show();
        }
        public void Start()
        {            
            // Ponemos en marcha el timer para capturar las imágenes
            // We start the timer to capture the images
            _bCapturing = true;
            _tCapture = new Thread(new ParameterizedThreadStart(CaptureThread));
            _tCapture.Start(_nTimerPeriod);
        }
        // Dejamos de capturar imágenes
        // We stop capturing images
        public void Close()
        {
            try
            {
                if (_tCapture != null)
                {
                    _bCapturing = false;
                    _tCapture.Abort();
                    _tCapture = null;
                }
            }
            catch
            {
            }
        }
        #endregion

        private string RequestCommand(string url, int timeout = 2000)
        {
            HttpWebRequest wrq = (HttpWebRequest)HttpWebRequest.Create(url);
            wrq.Method = "GET";
            // Proprciona un timeot corto a las operaciones de control para dar prioridad a la captura de imágenes
            // Provide a short timeot to the control operations to priorize the image capture
            wrq.Timeout = timeout;
            // Autenticación básica por encabezados http
            // Http headers basic Authentication
            wrq.Credentials = new NetworkCredential(UserName, Password);
            string response;
            WebResponse wr = wrq.GetResponse();
            Stream receiveStream = wr.GetResponseStream();
            StreamReader reader = new StreamReader(receiveStream, Encoding.UTF8);
            response = reader.ReadToEnd();
            wr.Close();
            return response;
        }
        private void ChangeResolution()
        {
            if (_nChangeResolution)
            {
                try
                {
                    string url = Uri + (Uri.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=list&group=Properties.Image.Resolution";
                    string[] reso = RequestCommand(url).Split('=')[1].Split(',');
                    int rres = int.Parse(_width + _height);
                    int ores = 0;
                    string fres = "";
                    foreach (string r in reso)
                    {
                        string[] tres = r.Split('x');
                        int test = int.Parse(tres[0] + tres[1]);
                        if (test == rres)
                        {
                            url = Uri + (Uri.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=update&Image.I0.Appearance.Resolution=" + r;
                            RequestCommand(url);
                            return;
                        }
                        else if ((test < rres) && (test > ores))
                        {
                            ores = test;
                            fres = r;
                        }
                    }
                    if (!string.IsNullOrEmpty(fres))
                    {
                        url = Uri + (Uri.EndsWith("/") ? "" : "/") + "axis-cgi/param.cgi?action=update&Image.I0.Appearance.Resolution=" + fres;
                        RequestCommand(url);
                    }
                }
                catch
                {
                }
                finally
                {
                    _nChangeResolution = false;
                }
            }
        }
        private void CaptureThread(object interval)
        {
            while (_bCapturing)
            {
                try
                {
                    DateTime tm = DateTime.Now;
                    // Comando para capturar una imágen
                    // Command to capture an image
                    string url = Uri + (Uri.EndsWith("/") ? "" : "/") + "axis-cgi/jpg/image.cgi";
                    HttpWebRequest wrq = (HttpWebRequest)HttpWebRequest.Create(url);
                    wrq.Method = "GET";
                    wrq.Credentials = new NetworkCredential(UserName, Password);
                    Bitmap bmp = null;
                    WebResponse wr = wrq.GetResponse();
                    // Recibimos un stream con una imagen en formato JPG
                    // We receive a stream with an image in JPG format
                    Stream receiveStream = wr.GetResponseStream();
                    bmp = new Bitmap(receiveStream);
                    wr.Close();
                    // Y la convertimos en un bitmap
                    // Por último, la pasamos al programa para que la procese
                    // And we turn it into a bitmap
                    // Finally, we provide the image to the program to be processed
                    OnNewFrame?.Invoke(this, new FrameEventArgs(bmp));
                    TimeSpan ts;
                    do
                    {
                        ts = DateTime.Now.Subtract(tm);
                    } while (((ts.Seconds * 1000) + ts.Milliseconds) <= (int)interval);
                }
                catch
                {
                }
            }
        }
        public override string ToString()
        {
            // Texto que se mostrará en la lista de cámaras de la galería de tiro
            // Text to be displayed in the camera list of the shooting gallery
            return Resources.ID_PROTOCOL;
        }
    }
}
