﻿using System.Collections.Generic;
using System.IO;
using WatcherCommons.Data;

namespace WatcherCommons.Interfaces
{
    /// <summary>
    /// Protocol to interchange files and messages between the clients and the server
    /// </summary>
    public interface IStorageManager
    {
        /// <summary>
        /// string with the protocol settings
        /// </summary>
        string ConnsecionString { get; set; }
        /// <summary>
        /// The file locator
        /// </summary>
        string ContainerPath { get; set; }
        /// <summary>
        /// Upload a file to the storage
        /// </summary>
        /// <param name="filename">Destination file name</param>
        /// <param name="s">Stream containing the file content</param>
        void UploadFile(string filename, Stream s);
        /// <summary>
        /// Download a file from the storage
        /// </summary>
        /// <param name="filename">Name of the file</param>
        /// <param name="s">Stream to receive the file content</param>
        void DownloadFile(string filename, Stream s);
        /// <summary>
        /// Delete a file from the storage
        /// </summary>
        /// <param name="filename">Name of the file to delete</param>
        void DeleteFile(string filename);
        /// <summary>
        /// Test if a given file exists in the storage
        /// </summary>
        /// <param name="filename">Name of the file</param>
        /// <returns></returns>
        bool ExistsFile(string filename);
        /// <summary>
        /// Retrieve a list of file names from the storage
        /// </summary>
        /// <param name="model">Model that file names must match</param>
        /// <returns>List of strings with the file names</returns>
        IEnumerable<string> ListFiles(string model);
        /// <summary>
        /// Get commands sent by the clients
        /// </summary>
        /// <returns>ControlCommand enumerator</returns>
        IEnumerable<ControlCommand> GetCommands();
        /// <summary>
        /// Get camera requests sent by the clients
        /// </summary>
        /// <returns>Each request is a list of CameraInfo data, each item for a different camera</returns>
        IEnumerable<List<CameraInfo>> GetRequests();
        /// <summary>
        /// Send response for a command or request to the client
        /// </summary>
        /// <param name="resp">A list of CameraInfo objects, each one for a camera</param>
        void SendResponse(List<CameraInfo> resp);
    }
}
