const Config = require('./config.js');
const AdapterClient = require('./client.js');
const WebServer = require('./server.js');
const SmartHome = require('./smarthome.js');
const Skills = require('./skills.js');
const Promise = require('promise');
const helpers = require('./helpers.js');

class Bot {
  constructor (env, logLevel) {
    const userData = Config.loadUser(env);
    const config = Config.loadGlobal(env);
    this.client = new AdapterClient(config.adapter);
    this.server = new WebServer(config.webserver);
    this.shc = new SmartHome(config.smarthome);
    this.skills = new Skills(this.shc);

    this.skills.on('success', (id, message) => {
      this.client.send('answer', id, message);
    });

    this.skills.on('failure', (id) => {
      this.client.send('answer', id, 'Sorry I did not understand your query. Please rephrase!');
    })

    this.client.on('query', (data) => {
      const intent = helpers.getIntent(data.intents);
      const entities = helpers.getEntities(data.entities);

      if (logLevel > 0) {
        console.log('Received %s with intent %s and entities %s', data.raw, intent, entities.join(', ')); 
      }

      this.skills.resolve(data.id, intent, entities);
    });

    this.client.on('connect', (host) => {
      console.log('WebSocket connected to %s!', host);

      if (userData.refreshToken) {
        recover(userData.refreshToken, userData.accounts || []);
      }
    });

    this.client.on('reconnect', (host) => {
      console.log('WebSocket reconnected to %s!', host);
    });

    this.client.on('disconnect', (host) => {
      console.log('WebSocket disconnected from %s!', host);
    });

    this.client.on('linked', (account) => {
      if (userData.accounts === undefined) {
        userData.accounts = [];
      }

      userData.accounts.push(account);
      Config.saveUser(env, userData);
    });

    this.server.on('index', (model) => {
      model.title = 'Welcome';
      model.client = this.client;
      model.config = config;
      model.user = userData;
      model.message = this.client.setup ? 
        'Local chat bot configuration overview' : 
        'Authorize your local chatbot to get started!';
    });

    this.server.on('setup', (model) => {
      const data = model.data;
      model.title = 'Setup Done';
      model.message = 'Please wait until the bot is fully connected.';

      this.shc.login(data.code).then(setup);
    });

    this.server.on('revoke', (model) => {
      const data = model.data;

      if (userData.accounts === undefined) {
        userData.accounts = [];
      }

      for (var i = userData.accounts.length - 1; i >= 0; i--) {
        var account = userData.accounts[i];

        if (account.provider === data.provider && account.id === data.id) {
          userData.accounts.splice(i, 1);
          Config.saveUser(env, userData);
          this.client.send('unlink', data.provider, data.id);
          break;
        }
      };
    });

    var setup = (data) => {
      userData.refreshToken = data.refresh;
      userData.expires = data.expires;
      Config.saveUser(env, userData);
      return this.client.configure(data.shc.serialnumber);
    }

    var recover = (refreshToken, accounts) => {
      this.shc.refresh(refreshToken).then(setup).then(() => {
        return Promise.all(accounts.map((account) => {
          return this.client.send('link', account.provider, account.id);
        }));
      });
    }
  }

  start () {
    return this.client.connect();
  }
};

module.exports = Bot;