const WebSocket = require('ws');
const Promise = require('promise');
const util = require('util');
const crypto = require('crypto');
const EventEmitter = require('events').EventEmitter;

function addSlashes (str) {
  return (str + '').replace(/[\\"']/g, '\\$&').replace(/\u0000/g, '\\0');
}

function generateToken (length) {
  return crypto.randomBytes(length).toString('hex');
}

class AdapterClient extends EventEmitter {
  constructor (config) {
    super();
    this.id = undefined;
    this.ws = undefined;
    this.setup = undefined;
    this.connected = false;
    this.address = 'ws://' + config.host + '/api/channel';
    this.handlers = {
      open: () => {
        this.connected = true;

        if (this.setup) {
          this.send('register', this.setup.serial, this.setup.token);
          this.emit('reconnect', config.host);
        } else {
          this.emit('connect', config.host);
        }
      },
      error: (err) => {
        console.log('(WebSocket) %s.', err);
      },
      connected: (id) => {
        this.id = id;
      },
      link: (data) => {
        const token = generateToken(16);
        this.setup.token = token;
        this.emit('linked', {
          provider: data.provider,
          id: data.account
        });
        this.send('update', token);
      },
      question: (request) => {      
        const evaluation = request.evaluation;

        if (this.setup) {
          this.emit('query', {
            id: request.id,
            raw: evaluation.query,
            intents: evaluation.intents,
            entities: evaluation.entities 
          });
        } else {
          this.send('answer', 
            request.id, 
            'Client not connected yet. Please configure the local client first.');
        }
      },
      close: () => {
        this.connected = false;
        this.emit('disconnect', config.host);
      },
    };

    setInterval(() => {
      if (this.ws && !this.connected) {
        this.connect();
      }
    }, 10 * 1000);
  }

  connect () {
    const ws = new WebSocket(this.address);
    ws.on('error', this.handlers.error);
    ws.on('open', this.handlers.open);    
    ws.on('close', this.handlers.close);
    ws.on('message', (message, flags) => {
      const data = JSON.parse(message);
      const handler = this.handlers[data.type];
      handler(data.value);
    });
    this.ws = ws;
  }

  configure (serial) {
    const token = generateToken(16);
    this.setup = {
      serial: serial,
      token: token
    };

    return this.send('register', serial).then(() => {
      return this.send('update', token);
    });
  }

  disconnect () {
    if (this.ws) {
      this.ws.close();
      this.ws = undefined;
    }
  }

  send (fn) {
    const args = Array.prototype.slice.call(arguments, 1).map(addSlashes);
    args.splice(0, 0, this.id);
    return new Promise((resolve, reject) => {
      this.ws.send(fn + '("' + args.join('", "') + '")', (err) => {
        if (err) {
          reject(err);
        } else {
          resolve();
        }
      });
    });
  };
};

util.inherits(AdapterClient, EventEmitter);
module.exports = AdapterClient;